// Copyright (c) Odile B
// Distributed under the terms of the Modified BSD License.

const styles = require('./sage-combinat-widgets.css');
export * from './version';
export * from './singleton_widgets';
